'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 01 - Triângulo
'''
from OpenGL.GL import *
from OpenGL.GL import shaders
import numpy as np
import os
import pygame as pg
import sys


class OpenGLWindow:

    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec2 position;
layout (location = 1) in vec3 color;
out vec3 vertex_color;
void main()
{
    gl_Position = vec4(position, 0.0, 1.0);
    vertex_color = color;
}'''

    FRAGMENT_SHADER = '''#version 330 core
in vec3 vertex_color;
out vec4 color;
void main()
{
    color = vec4(vertex_color, 1.0f);
}'''
    
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        #                    Posições    Cores
        vertices = np.array([-0.5, -0.5, 1.0, 0.0, 0.0, # Esquerdo
                              0.5, -0.5, 0.0, 1.0, 0.0, # Direito
                              0.0,  0.5, 0.0, 0.0, 1.0  # Superior
                        ], dtype=np.float32)
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria um vertex buffer object
        vbo = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, vertices.size * vertices.itemsize,
                     vertices, GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y
        glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 5 * vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Especifica o local das cores r, g, b 
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 5 * vertices.itemsize,
                              ctypes.c_void_p(2 * vertices.itemsize))
        glEnableVertexAttribArray(1)
        
        # Desconecta vao
        glBindVertexArray(0)
        
    def display(self):
        # Define a cor de fundo da janela
        glClearColor(0.2, 0.3, 0.3, 1.0)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT)
        
        # Desenha o triângulo
        glUseProgram(self.program)
        glBindVertexArray(self.vao)
        glDrawArrays(GL_TRIANGLES, 0, 3)
        glBindVertexArray(0)
        glUseProgram(0)
    
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Triângulo')
    window.show()


if __name__ == '__main__':
    main()