'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 04 - Cubo
'''
from OpenGL.GL import *
from OpenGL.GL import shaders
import numpy as np
import os
import pygame as pg
import pyrr
import sys


class OpenGLWindow:

    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec3 position;
layout (location = 1) in vec3 color;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

out vec3 vertex_color;

void main()
{
    gl_Position = projection * view * model * vec4(position, 1.0f);
    vertex_color = color;
}'''

    FRAGMENT_SHADER = '''#version 330 core
in vec3 vertex_color;
out vec4 color;
void main()
{
    color = vec4(vertex_color, 1.0f);
}'''
    
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        vertices = np.array([
                             # Frente          Cores
                             -0.5,  0.5,  0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5,  0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                              0.5, -0.5,  0.5, 1.0, 1.0, 1.0,
                             
                             # Fundo           Cores
                             -0.5,  0.5, -0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5, -0.5, 0.0, 0.0, 1.0,
                              0.5, -0.5, -0.5, 1.0, 1.0, 1.0,
                              
                             # Esquerda        Cores
                             -0.5,  0.5,  0.5, 1.0, 0.0, 0.0,
                             -0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                             -0.5, -0.5, -0.5, 1.0, 1.0, 1.0,
                              
                             # Direita         Cores
                              0.5,  0.5,  0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                             -0.5, -0.5, -0.5, 1.0, 1.0, 1.0,
                             
                             # Topo            Cores
                             -0.5,  0.5, -0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5,  0.5,  0.5, 0.0, 0.0, 1.0,
                              0.5,  0.5,  0.5, 1.0, 1.0, 1.0,
                             
                             # Base            Cores
                             -0.5, -0.5, -0.5, 1.0, 0.0, 0.0,
                              0.5, -0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                              0.5, -0.5,  0.5, 1.0, 1.0, 1.0
                            ], dtype=np.float32)
        
        self.indices = np.array([
                                 0,  1,  2,  1,  2,  3,
                                 4,  5,  6,  5,  6,  7,
                                 8,  9, 10,  9, 10, 11,
                                12, 13, 14, 13, 14, 15,
                                16, 17, 18, 17, 18, 19,
                                20, 21, 22, 21, 22, 23 
                               ], dtype=np.uint8)
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria um vertex buffer object
        vbo = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, vertices.size * vertices.itemsize,
                     vertices, GL_STATIC_DRAW)
        
        # Cria um element buffer object
        ebo = glGenBuffers(1)
        # Conecta objeto ebo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        # Copia dados na memória
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.indices.size * self.indices.itemsize,
                     self.indices, GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y, z
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Especifica o local das cores r, g, b 
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * vertices.itemsize,
                              ctypes.c_void_p(3 * vertices.itemsize))
        glEnableVertexAttribArray(1)
        
        # Desconecta vao
        glBindVertexArray(0)
        
        # Cria matriz de modelo
        model = pyrr.Matrix44.from_scale(scale=(1.1, 1.1, 1.1), dtype=np.float16)
        model *= pyrr.Matrix44.from_x_rotation(np.radians(-45.), dtype=np.float16)
        model *= pyrr.Matrix44.from_y_rotation(np.radians(-45.), dtype=np.float16)
        model = model.astype(np.float32)
        
        # Cria matriz de visualização
        view = pyrr.Matrix44.look_at(pyrr.Vector3([0., 0., 3.], dtype=np.float16),
                                     pyrr.Vector3([0., 0., 0.], dtype=np.float16),
                                     pyrr.Vector3([0., 1., 0.], dtype=np.float16),
                                     dtype=np.float32)
        
        # Cria matriz de projeção
        projection = pyrr.Matrix44.perspective_projection(45.,
                                                          self.width / self.height,
                                                          .1,
                                                          100.,
                                                          dtype=np.float32)
        
        glUseProgram(self.program)
        
        # Retorna a localização das variáveis uniformes
        model_loc = glGetUniformLocation(self.program, 'model')
        view_loc = glGetUniformLocation(self.program, 'view')
        proj_loc = glGetUniformLocation(self.program, 'projection')
        
        # Define o valor das variáveis uniformes
        glUniformMatrix4fv(view_loc, 1, GL_FALSE, view)
        glUniformMatrix4fv(model_loc, 1, GL_FALSE, model)
        glUniformMatrix4fv(proj_loc, 1, GL_FALSE, projection)
        
        glUseProgram(0)
        
        # Teste de profundidade (utilizar em 3D)
        glEnable(GL_DEPTH_TEST)
    
    def display(self):
        # Define a cor de fundo da janela
        glClearColor(0.2, 0.3, 0.3, 1.0)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        # Desenha o cubo
        glUseProgram(self.program)
        glBindVertexArray(self.vao)
        glDrawElements(GL_TRIANGLES, self.indices.size, GL_UNSIGNED_BYTE, None)
        glBindVertexArray(0)
        glUseProgram(0)
    
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Cubo')
    window.show()


if __name__ == '__main__':
    main()