'''
Instituto Tecnológico de Aeronáutica
CCI 36 – Computação Gráfica - 2018
Prof. Carlos Henrique Q. Forster
Luiz Felipe Simões Hoffmann

Exemplo 05 - Move Cubos

Utilizar as teclas RIGHT, LEFT, DOWN, UP, para mover os cubos.
'''
from OpenGL.GL import *
from OpenGL.GL import shaders
import numpy as np
import os
import pygame as pg
import pyrr
import sys


class OpenGLWindow:

    VERTEX_SHADER = '''#version 330 core
layout (location = 0) in vec3 position;
layout (location = 1) in vec3 color;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;

out vec3 vertex_color;

void main()
{
    gl_Position = projection * view * model * vec4(position, 1.0f);
    vertex_color = color;
}'''

    FRAGMENT_SHADER = '''#version 330 core
in vec3 vertex_color;
out vec4 color;
void main()
{
    color = vec4(vertex_color, 1.0f);
}'''
    
    def __init__(self, width, height, title):
        self.width = width
        self.height = height
        
        self.camera_pos = pyrr.Vector3([0.0, 0.0, 3.0], dtype=np.float16)
        self.camera_front = pyrr.Vector3([0.0, 0.0, -1.0], dtype=np.float16)
        self.camera_up = pyrr.Vector3([0.0, 1.0, 0.0], dtype=np.float16)
        
        self.delta_time = 0.0
        self.last_frame = 0.0
        
        self.keys = {pg.K_RIGHT: False,
                     pg.K_LEFT: False,
                     pg.K_DOWN: False,
                     pg.K_UP: False}
        
        os.environ['SDL_VIDEO_CENTERED'] = '1'
        
        # Inicia o pygame
        pg.init()
        
        self.clock = pg.time.Clock()
        
        # Cria a janela
        pg.display.set_caption(title)
        screen = pg.display.set_mode((width, height),
                                     pg.HWSURFACE | pg.OPENGL | pg.DOUBLEBUF)
        self.resize(width, height)
        
        self.init()
    
    def init(self):
        # Compila os shaders
        vertex_shader = shaders.compileShader(self.VERTEX_SHADER, GL_VERTEX_SHADER)
        fragment_shader = shaders.compileShader(self.FRAGMENT_SHADER, GL_FRAGMENT_SHADER)
        self.program = shaders.compileProgram(vertex_shader, fragment_shader)
        
        glDeleteShader(vertex_shader)
        glDeleteShader(fragment_shader)
        
        vertices = np.array([
                             # Frente          Cores
                             -0.5,  0.5,  0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5,  0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                              0.5, -0.5,  0.5, 1.0, 1.0, 1.0,
                             
                             # Fundo           Cores
                             -0.5,  0.5, -0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5, -0.5, 0.0, 0.0, 1.0,
                              0.5, -0.5, -0.5, 1.0, 1.0, 1.0,
                              
                             # Esquerda        Cores
                             -0.5,  0.5,  0.5, 1.0, 0.0, 0.0,
                             -0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                             -0.5, -0.5, -0.5, 1.0, 1.0, 1.0,
                              
                             # Direita         Cores
                              0.5,  0.5,  0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                             -0.5, -0.5, -0.5, 1.0, 1.0, 1.0,
                             
                             # Topo            Cores
                             -0.5,  0.5, -0.5, 1.0, 0.0, 0.0,
                              0.5,  0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5,  0.5,  0.5, 0.0, 0.0, 1.0,
                              0.5,  0.5,  0.5, 1.0, 1.0, 1.0,
                             
                             # Base            Cores
                             -0.5, -0.5, -0.5, 1.0, 0.0, 0.0,
                              0.5, -0.5, -0.5, 0.0, 1.0, 0.0,
                             -0.5, -0.5,  0.5, 0.0, 0.0, 1.0,
                              0.5, -0.5,  0.5, 1.0, 1.0, 1.0
                            ], dtype=np.float32)
        
        self.indices = np.array([
                                 0,  1,  2,  1,  2,  3,
                                 4,  5,  6,  5,  6,  7,
                                 8,  9, 10,  9, 10, 11,
                                12, 13, 14, 13, 14, 15,
                                16, 17, 18, 17, 18, 19,
                                20, 21, 22, 21, 22, 23 
                               ], dtype=np.uint8)
        
        self.cube_positions = [pyrr.Vector3([ 0.0,  0.0,   0.0]),
                               pyrr.Vector3([ 2.0,  5.0, -15.0]),
                               pyrr.Vector3([-1.5, -2.2, - 2.5]),
                               pyrr.Vector3([-3.8, -2.0, -12.3]),
                               pyrr.Vector3([ 2.4, -0.4, - 3.5]),
                               pyrr.Vector3([-1.7,  3.0, - 7.5]),
                               pyrr.Vector3([ 1.3, -2.0, - 2.5]),
                               pyrr.Vector3([ 1.5,  2.0, - 2.5]),
                               pyrr.Vector3([ 1.5,  0.2, - 1.5]),
                               pyrr.Vector3([-1.3,  1.0, - 1.5])]
        
        # Cria um vertex array object
        self.vao = glGenVertexArrays(1)
        # Conecta objeto vao
        glBindVertexArray(self.vao)
    
        # Cria um vertex buffer object
        vbo = glGenBuffers(1)
        # Conecta objeto vbo
        glBindBuffer(GL_ARRAY_BUFFER, vbo)
        # Copia dados na memória
        glBufferData(GL_ARRAY_BUFFER, vertices.size * vertices.itemsize,
                     vertices, GL_STATIC_DRAW)
        
        # Cria um element buffer object
        ebo = glGenBuffers(1)
        # Conecta objeto ebo
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ebo)
        # Copia dados na memória
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, self.indices.size * self.indices.itemsize,
                     self.indices, GL_STATIC_DRAW)
        
        # Especifica o local das coordenadas (posições) x, y, z
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 6 * vertices.itemsize, None)
        glEnableVertexAttribArray(0)
        
        # Especifica o local das cores r, g, b 
        glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 6 * vertices.itemsize,
                              ctypes.c_void_p(3 * vertices.itemsize))
        glEnableVertexAttribArray(1)
        
        # Desconecta vao
        glBindVertexArray(0)
        
        # Cria matriz de projeção
        projection = pyrr.Matrix44.perspective_projection(45.,
                                                          self.width / self.height,
                                                          .1,
                                                          100.,
                                                          dtype=np.float32)
        
        glUseProgram(self.program)
        
        # Retorna a localização das variáveis uniformes
        self.model_loc = glGetUniformLocation(self.program, 'model')
        self.view_loc = glGetUniformLocation(self.program, 'view')
        proj_loc = glGetUniformLocation(self.program, 'projection')
        
        # Define o valor da variável uniforme `projection`
        glUniformMatrix4fv(proj_loc, 1, GL_FALSE, projection)
        
        glUseProgram(0)
        
        # Teste de profundidade (utilizar em 3D)
        glEnable(GL_DEPTH_TEST)
    
    def display(self):
        current_frame = pg.time.get_ticks() / 1000
        self.delta_time = current_frame - self.last_frame
        self.last_frame = current_frame
        self.do_movement()
        
        # Define a cor de fundo da janela
        glClearColor(0.2, 0.3, 0.3, 1.0)
        
        # Limpa o buffer da janela
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
        
        glUseProgram(self.program)
        glBindVertexArray(self.vao)
        
        # Cria matriz de visualização
        view = pyrr.Matrix44.look_at(self.camera_pos,
                                     self.camera_pos + self.camera_front,
                                     self.camera_up,
                                     dtype=np.float32)
        # Define o valor da variável uniforme `view`
        glUniformMatrix4fv(self.view_loc, 1, GL_FALSE, view)
        
        # Desenha os cubos
        for i in range(len(self.cube_positions)):
            angle = 20.0 * i
            # Cria matriz de modelo
            model = pyrr.Matrix44.from_translation(self.cube_positions[i], dtype=np.float16)
            model *= pyrr.matrix44.create_from_axis_rotation(pyrr.Vector3([1.0, 0.3, 0.5]), np.radians(angle), dtype=np.float16)
            model = model.astype(np.float32)
            # Define o valor da variável uniforme `model`
            glUniformMatrix4fv(self.model_loc, 1, GL_FALSE, model)
            glDrawElements(GL_TRIANGLES, self.indices.size, GL_UNSIGNED_BYTE, None)
        
        glBindVertexArray(0)
        glUseProgram(0)
    
    def do_movement(self):
        speed = 5.0 * self.delta_time
        
        if (self.keys[pg.K_UP]):
            self.camera_pos += self.camera_front * speed
        if (self.keys[pg.K_DOWN]):
            self.camera_pos -= self.camera_front * speed
        if (self.keys[pg.K_LEFT]):
            self.camera_pos -= self.camera_front.cross(self.camera_up).normalized * speed 
        if (self.keys[pg.K_RIGHT]):
            self.camera_pos += self.camera_front.cross(self.camera_up).normalized * speed
    
    def resize(self, width, height):
        # Define o tamanho da janela de desenho
        glViewport(0, 0, width, height)
        
    def show(self):
        # Permanece em loop ate o usuário fechar a janela
        while True:
            for event in pg.event.get():
                if event.type == pg.QUIT:
                    pg.quit()
                    sys.exit()
                elif (event.type == pg.KEYDOWN and event.key in self.keys):
                    self.keys[event.key] = True
                elif (event.type == pg.KEYUP and event.key in self.keys):
                    self.keys[event.key] = False
            
            #Renderiza
            self.display()
            
            # Atualiza a janela
            pg.display.flip()
            
            # Define a velocidade com que a janela é atualizada
            self.clock.tick(65)


def main():
    window = OpenGLWindow(800, 600, 'CCI-36 - Move Cubo')
    window.show()


if __name__ == '__main__':
    main()